package com.streambase.contrib.tuples;

/*
 * @Author Adam Diamant
 */

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.NullValueException;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.Schema.Field;
import com.streambase.sb.client.CustomFunctionResolver;

public class MortgageCalculator {
	
	/**
	 * This function rounds the decimal values to only two decimal points
	 * @param number The number one wishes to round
	 * @param places The number of decimal places to round the number off
	 * @return A decimal number with only two decimal places 
	 */
	public static double Round(double number, int places) {
		  double p = (double)Math.pow(10,places);
		  number = number * p;
		  return (double) Math.round(number)/p;
	}
	
	/**
	 * This function calculates mortgage payments and then puts the data into a tuple
	 * @param Name 		The name of the individual with the mortgage
	 * @param Amount  	The total amount of the loan taken out
	 * @param Term		The life of the mortgage in months
	 * @param Interest	The yearly interest rate
	 * @param Months	The numbers of the months of extra payments
	 * @param Values	The amount of extra-payment for each of the extra-payment months
	 * @return A tuple representing all the relevant data associated with this mortgage 
	 */	
	@SuppressWarnings("unchecked")
	@CustomFunctionResolver("MakePaymentsResolver")
	public static Tuple payments(String Name, double Amount, int Term, double Interest, List<?> Months, List<?> Values){
		if (Months.size() != Values.size()){
			System.err.println("The length of the list of extra payment months and their corresponding values is not the same");
			return null;
		}
		if (Amount <= 0 || Term <= 0 || Interest <= 0 || Interest > 100.0){
			System.err.println("One or more of the input parameters are incorrect");
			return null;
		}
		
		// Declaring the lists to return
		ArrayList<Double> principal = new ArrayList<Double>();
		ArrayList<Double> total = new ArrayList<Double>();
		ArrayList<Double> interest = new ArrayList<Double>();
		ArrayList<Double> payments = new ArrayList<Double>();
		
		// Declaring the variables used in the calculation
		double r = Interest/(100*12.0), tInterest = 0, sum, prepayment; 
		double balanceCur = Amount, balancePrev = Amount, payment; 
		int month = -1, Nminus1,counter = 0;
		
		// Creating fields to add to the tuple
		Field[] f = new Field[10];
		f[0] = Schema.createField(DataType.STRING, "Name");			
		f[1] = Schema.createField(DataType.DOUBLE, "MortgageAmount");		
		f[2] = Schema.createField(DataType.INT, "MortgageTerm");		
		f[3] = Schema.createField(DataType.DOUBLE, "InterestRate");				
		f[4] = Schema.createField(DataType.DOUBLE, "InterestPaid");		
		f[5] = Schema.createField(DataType.DOUBLE, "TotalAmount");		
		f[6] = Schema.createListField("PrincipalPayments", CompleteDataType.forDouble());		
		f[7] = Schema.createListField("InterestPayments", CompleteDataType.forDouble());		
		f[8] = Schema.createListField("TotalOutstanding", CompleteDataType.forDouble());		
		f[9] = Schema.createListField("MonthlyPayments",CompleteDataType.forDouble());
		
		
		// Set Descriptions
		f[0].setDescription("The name of the party taking out the mortgage");
		f[1].setDescription("The principal amount of the mortgage");
		f[2].setDescription("The term (in months) of the mortgage (ie. 360 months represents a mortgage of 30 years)");
		f[3].setDescription("The annualized simple interest interest rate as a percentage");
		f[4].setDescription("The amount of of money paid towards interest rate payments only");
		f[5].setDescription("The total amount of money (interest rate payments + principal) used to pay off the mortgage");
		f[6].setDescription("A list of values representing the amount of principal paid each month (over the life of the mortgage)");
		f[7].setDescription("A list of values representing the amount of interest paid each month (over the life of the mortgage)");
		f[8].setDescription("A list of values representing the total amount outstanding of the mortgage from month to month");
		f[9].setDescription("A list of values representing the monthly payments made towards the mortgage (including the pre-payments)");
		
		// Creating a schema
		Schema schema = new Schema(null, f);
		
		// Creating a tuple from the schema above
		Tuple tuple = schema.createTuple();	
		
		// Calculating the mortgage statistics
		total.add(Round(Amount,2));	
		payment = (r * Math.pow(1+r,Term) * Amount)/(Math.pow(1+r,Term)- 1);
		Iterator<Integer> m = (Iterator<Integer>) Months.iterator();
		Iterator<Double> v = (Iterator<Double>) Values.iterator();
		
		// Calculate the next month of pre-payment
		if (m.hasNext())
			month = m.next().intValue();
		
		// If the pre-payment months inputed are not within the mortgage time period
		// then the mortgage pre-payment is not computed
		if (month <= 0 || month >= Term)
			month = -1;
	
		// Simulate until the balance of the mortgage is paid off
		while (balanceCur >= 0){		
			counter++;
			balancePrev = balanceCur;		
			
			// Calculate outstanding mortgage amount 
			balanceCur = balanceCur*(1+r)- payment;	
			
			// Append values to their appropriate lists
			total.add(Round(Math.max(balanceCur, 0),2));	
			interest.add(Round(r*balancePrev,2));
			principal.add(Round(payment - r*balancePrev,2));
			payments.add(Round(payment,2));
			
			// Reduce balance after pre-payment months
			if (counter == month){
				prepayment = v.next().doubleValue();
				// Prepayment cannot be negative or more then the balance of the mortgage
				if (prepayment > 0 && prepayment <= balanceCur){
					payments.set(counter-1, payments.get(counter-1) + prepayment); 
					balanceCur = balanceCur - prepayment;	
				}
				// The prepayment months cannot be greater then the end of the term nor can 
				// they be negative. Note that if the program exists before a pre-payment can
				// occur, it obviously does not contribute to the final statistics.
				if (m.hasNext())
					month = m.next().intValue();
				if (month <= 0 || month >= Term)
					month = -1;
			}
		}			
		
		// Append final values to make all lists equal in length		
		interest.add(0.0);
		principal.add(Round(payment,2));
		payments.add(Round(balancePrev,2));

				
		// Calculate the total amount of interest paid
		Nminus1 = principal.size()-1;
		for (int i = 0; i < Nminus1; i++){
			tInterest += interest.get(i);
		}
		
		// Calculate the total amount of money paid
		sum = Amount + tInterest;

		// Write the values obtained to the tuple
		try {
			tuple.setString("Name", Name);
			tuple.setDouble("MortgageAmount", Amount);
			tuple.setInt("MortgageTerm", counter);
			tuple.setDouble("InterestRate", Interest);	
			tuple.setDouble("InterestPaid", Round(tInterest,2));
			tuple.setDouble("TotalAmount",Round(sum,2));
			tuple.setList("PrincipalPayments", principal);
			tuple.setList("InterestPayments", interest);
			tuple.setList("TotalOutstanding", total);
			tuple.setList("MonthlyPayments",payments);
		} catch (NullValueException e) {
			System.err.println("Null value passed into a tuple field");
			e.printStackTrace();
		} catch (TupleException e) {
			System.err.println("An error was generated creating a tuple");
			e.printStackTrace();
		}
		return tuple;
	}
	
	/**
	 * A StreamBase Custom Function Resolver Function. This method is used by 
	 * StreamBase for type checking.  Use this method
	 * to resolve functions that take lists or tuples as arguments 
	 * or return lists or tuples.  This method should be used to ensure the 
	 * data types of your function are correct.
	 * @throws StreamBaseException 
	*/
	public static CompleteDataType MakePaymentsResolver(
			CompleteDataType Name, CompleteDataType Amount, CompleteDataType Term, CompleteDataType Interest,
			CompleteDataType Months, CompleteDataType Values) throws StreamBaseException {
		CompleteDataType months = Months.getElementType();
		CompleteDataType values = Values.getElementType();
		if(Name.getDataType() == DataType.STRING && Amount.getDataType() == DataType.DOUBLE && 
				Term.getDataType() == DataType.INT && Interest.getDataType() == DataType.DOUBLE && 
				months.getDataType() == DataType.INT && values.getDataType() == DataType.DOUBLE ){
			
			// Set fields
			Field[] f = new Field[10];
			f[0] = Schema.createField(Name.getDataType(), "Name");
			f[1] = Schema.createField(Amount.getDataType(), "MortgageAmount");
			f[2] = Schema.createField(Term.getDataType(), "MortgageTerm");
			f[3] = Schema.createField(Interest.getDataType(), "InterestRate");
			f[3] = Schema.createField(DataType.DOUBLE, "InterestRate");
			f[4] = Schema.createField(DataType.DOUBLE, "InterestPaid");
			f[5] = Schema.createField(DataType.DOUBLE,"TotalAmount");
			f[6] = Schema.createListField("PrincipalPayments", Values.getElementType());
			f[7] = Schema.createListField("InterestPayments", Values.getElementType());
			f[8] = Schema.createListField("TotalOutstanding", Values.getElementType());
			f[9] = Schema.createListField("MonthlyPayments", Values.getElementType());

			// Set Descriptions
			f[0].setDescription("The name of the party taking out the mortgage");
			f[1].setDescription("The principal amount of the mortgage");
			f[2].setDescription("The term (in months) of the mortgage (ie. 360 months represents a mortgage of 30 years)");
			f[3].setDescription("The annualized simple interest interest rate as a percentage");
			f[4].setDescription("The amount of of money paid towards interest rate payments only");
			f[5].setDescription("The total amount of money (interest rate payments + principal) used to pay off the mortgage");
			f[6].setDescription("A list of values representing the amount of principal paid each month (over the life of the mortgage)");
			f[7].setDescription("A list of values representing the amount of interest paid each month (over the life of the mortgage)");
			f[8].setDescription("A list of values representing the total amount outstanding of the mortgage from month to month");
			f[9].setDescription("A list of values representing the monthly payments made towards the mortgage (including the pre-payments)");
			
			// Create output schema
			Schema schema = new Schema(null, f);	
			return CompleteDataType.forTuple(schema);
		}
		return null;	
	}

}
